from __future__ import annotations
from typing import Callable, Type

from domain_model.aggregate import Aggregate
from domain_model.commands import CommandsStorage
from domain_model.events_handlers import DomainEventsHandler
from domain_model.passport.repository import PassportRepository
from domain_model.repository import BaseRepository, BaseStorage
from domain_model.typing import AggregateData
from domain_model.unit_of_work.unit_of_work import UnitOfWork


def _get_default_repository_factory(storage_class: Type[BaseStorage]) -> Callable[[Passport, UnitOfWork], PassportRepository]:
    def _factory(passport: Passport, unit_of_work: UnitOfWork) -> PassportRepository:
        return PassportRepository(passport=passport, storage=storage_class(unit_of_work=unit_of_work))
    return _factory


def _default_aggregate_factory(data: AggregateData, commands_storage: CommandsStorage) -> Aggregate:
    return Aggregate(commands_storage=commands_storage, data=data)


class Passport:
    def __init__(
        self,
        commands_storage: CommandsStorage,
        repository_factory: Callable[[Passport, UnitOfWork], BaseRepository] = None,
        storage_class: Type[BaseStorage] = None,
        aggregate_factory: Callable[[AggregateData, CommandsStorage], Aggregate] = None,
        events_handler: DomainEventsHandler = None,
    ):
        self._commands_storage = commands_storage

        self._aggregate_factory = aggregate_factory or _default_aggregate_factory
        self._events_handler = events_handler or DomainEventsHandler({})

        if all((repository_factory, storage_class)) or not any((repository_factory, storage_class)):
            raise ValueError('Either `repository_factory` or `storage_class` must be provided, but not both.')

        self._repository_factory = repository_factory or _get_default_repository_factory(storage_class=storage_class)
        self._repository: BaseRepository | None = None

    def get_repository(self, unit_of_work: UnitOfWork) -> BaseRepository:
        if self._repository is None:
            self._repository = self._repository_factory(self, unit_of_work)

        return self._repository

    def create_aggregate(self, data: AggregateData) -> Aggregate:
        return self._aggregate_factory(data, self._commands_storage)

    def get_events_handler(self) -> DomainEventsHandler:
        return self._events_handler



from abc import abstractmethod
from typing import Any, Generator
from uuid import UUID

from domain_model.commands import BaseCommand
from domain_model.domain_events import BaseDomainEvent
from domain_model.events_handlers import DomainEventsHandler
from domain_model.passport.passport import Passport
from domain_model.typing import AggregateData
from domain_model.unit_of_work.unit_of_work import UnitOfWork
from domain_model.use_case import BaseUseCase, convert_error


class BasePassportUseCase(BaseUseCase):
    def __init__(self, passport: Passport, error_class: type[Exception]):
        super().__init__(error_class=error_class)
        self._passport = passport

    async def execute(self, unit_of_work: UnitOfWork, root_id: UUID, raw_data: Any) -> Any:
        validated_data = await self._parse_and_validate(raw_data=raw_data)
        commands = list(self._generate_commands(validated_data=validated_data))

        if not commands:
            raise ValueError('No commands generated.')

        repo = self._passport.get_repository(unit_of_work=unit_of_work)

        with convert_error(self._error_class):
            aggregate = await repo.get(root_id=root_id)

        with convert_error(self._error_class):
            aggregate.handle_commands(commands=commands)

        await repo.save(aggregate=aggregate)

        aggregate_data = aggregate.get_data()
        events = aggregate.get_domain_events()
        await self._handle_events(aggregate_data=aggregate_data, events=events, unit_of_work=unit_of_work)

        return self._prepare_result(aggregate_data=aggregate_data, events=events)

    async def _parse_and_validate(self, raw_data: Any) -> Any:
        parsed_data = await self._parse(raw_data=raw_data)
        validated_data = await self._validate(parsed_data=parsed_data)
        return validated_data

    async def _parse(self, raw_data: Any) -> Any:
        return raw_data

    async def _validate(self, parsed_data: Any) -> Any:
        return parsed_data

    @abstractmethod
    def _generate_commands(self, validated_data: Any) -> Generator[BaseCommand, None, None]:
        raise NotImplementedError

    async def _handle_events(self, aggregate_data: AggregateData, events: list[BaseDomainEvent], unit_of_work: UnitOfWork) -> None:
        events_handler = self._get_events_handler()
        await events_handler.handle(data=aggregate_data, events=events, unit_of_work=unit_of_work)

    def _get_events_handler(self) -> DomainEventsHandler:
        return self._passport.get_events_handler()

    def _prepare_result(self, aggregate_data: AggregateData, events: list[BaseDomainEvent]) -> Any:
        return None


class BaseCreateAggregatePassportUseCase(BaseUseCase):
    def __init__(self, passport: Passport, error_class: type[Exception]):
        super().__init__(error_class=error_class)
        self._passport = passport

    async def execute(self, unit_of_work: UnitOfWork, raw_data: Any) -> Any:
        validated_data = await self._parse_and_validate(raw_data=raw_data)
        commands = list(self._generate_commands(validated_data=validated_data))

        if not commands:
            raise ValueError('No commands generated.')

        repo = self._passport.get_repository(unit_of_work=unit_of_work)

        with convert_error(self._error_class):
            aggregate = await repo.create()

        with convert_error(self._error_class):
            aggregate.handle_commands(commands=commands)

        await repo.save(aggregate=aggregate)

        aggregate_data = aggregate.get_data()
        events = aggregate.get_domain_events()
        await self._handle_events(aggregate_data=aggregate_data, events=events, unit_of_work=unit_of_work)

        return self._prepare_result(aggregate_data=aggregate_data, events=events)

    async def _parse_and_validate(self, raw_data: Any) -> Any:
        parsed_data = await self._parse(raw_data=raw_data)
        validated_data = await self._validate(parsed_data=parsed_data)
        return validated_data

    async def _parse(self, raw_data: Any) -> Any:
        return raw_data

    async def _validate(self, parsed_data: Any) -> Any:
        return parsed_data

    @abstractmethod
    def _generate_commands(self, validated_data: Any) -> Generator[BaseCommand, None, None]:
        raise NotImplementedError

    async def _handle_events(self, aggregate_data: AggregateData, events: list[BaseDomainEvent], unit_of_work: UnitOfWork) -> None:
        events_handler = self._get_events_handler()
        await events_handler.handle(data=aggregate_data, events=events, unit_of_work=unit_of_work)

    def _get_events_handler(self) -> DomainEventsHandler:
        return self._passport.get_events_handler()

    def _prepare_result(self, aggregate_data: AggregateData, events: list[BaseDomainEvent]) -> Any:
        return None



from contextlib import AsyncExitStack

from domain_model.unit_of_work.enums import UnitStatus
from domain_model.unit_of_work.typing import UnitTypeBase

from domain_model.unit_of_work.units.base import BaseUnit


class UnitOfWork:
    def __init__(
        self,
        units: dict[UnitTypeBase, BaseUnit],
    ):
        self._units = units
        self._status: UnitStatus = UnitStatus.READY

        self._units_statuses = {
            unit_type: UnitStatus.READY
            for unit_type, unit in self._units.items()
        }
        self._stack = AsyncExitStack()

    async def __aenter__(self):
        if self._status is not UnitStatus.READY:
            raise Exception('UnitOfWork could be entered only once.')

        self._status = UnitStatus.ENTERED
        # Units are lazy loaded when they are requested.

        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            for unit_type, unit in self._units.items():
                if self._units_statuses[unit_type] is UnitStatus.ENTERED:
                    await unit.handle_exception(exc_type, exc_val, exc_tb)

        await self._stack.aclose()

        self._status = UnitStatus.EXITED
        self._units_statuses = {
            unit_type: UnitStatus.EXITED
            for unit_type, unit in self._units.items()
        }

    async def get_unit(self, unit_type: UnitTypeBase) -> BaseUnit:
        if self._status is not UnitStatus.ENTERED:
            raise Exception('UnitOfWork is not entered.')

        if unit_type not in self._units:
            raise Exception(f'Unit type {unit_type} is not supported.')

        await self._enter_unit(unit_type)

        return self._units[unit_type]

    async def _enter_unit(self, unit_type: UnitTypeBase):
        if self._units_statuses[unit_type] is UnitStatus.ENTERED:
            return

        if self._units_statuses[unit_type] is UnitStatus.EXITED:
            raise Exception('Unit is already exited.')

        await self._stack.enter_async_context(self._units[unit_type])

        self._units_statuses[unit_type] = UnitStatus.ENTERED



from __future__ import annotations

from typing import Any
from uuid import UUID, uuid4

from pydantic import GetCoreSchemaHandler
from pydantic_core import core_schema


class ObjectID:
    def __init__(self, id_: UUID | None):
        self._initial_id = id_
        self._id = id_ or uuid4()

    @property
    def value(self) -> UUID:
        return self._id

    @property
    def is_new(self) -> bool:
        return self._initial_id is None

    def __eq__(self, other: ObjectID | UUID | None) -> bool:
        if isinstance(other, UUID):
            return self._id == other
        elif isinstance(other, ObjectID):
            return self._id == other._id
        # I'm not sure about this one - it's a bit unclear why comparison to `None` can return `True`.
        # However, it's quite useful in tests.
        elif other is None:
            return self._initial_id is None

        raise ValueError(f'Cannot compare ObjectID with {type(other)}')

    def __hash__(self) -> int:
        return hash(self._id)

    @classmethod
    def __get_pydantic_core_schema__(
            cls,
            _source_type: Any,
            _handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        return core_schema.uuid_schema()



from domain_model.aggregate import Aggregate
from domain_model.repository import BaseRepository, BaseStorage
from domain_model.typing import AggregateData


class PassportRepository(BaseRepository):
    def __init__(self, passport: 'Passport', storage: BaseStorage):
        super().__init__(storage=storage)
        self._passport = passport

    def _init_aggregate(self, data: AggregateData) -> Aggregate:
        return self._passport.create_aggregate(data=data)



from abc import ABC, abstractmethod
from uuid import UUID

from domain_model.aggregate import Aggregate
from domain_model.typing import AggregateData, AggregateRawData
from domain_model.unit_of_work.unit_of_work import UnitOfWork


class NotFoundError(Exception):
    def __init__(self, msg: str):
        msg = msg or 'Not found.'
        super().__init__(msg)
        self.message = msg


class BaseStorage(ABC):
    def __init__(self, unit_of_work: UnitOfWork):
        self._unit_of_work = unit_of_work

    @abstractmethod
    async def fetch_raw_data(self, root_id: UUID) -> AggregateRawData:
        pass

    @abstractmethod
    async def save(self, data: AggregateData) -> None:
        pass

    @abstractmethod
    def convert_raw_to_aggregate_data(self, raw_data: AggregateRawData) -> AggregateData:
        pass

    @abstractmethod
    async def initialize_new_raw_data(self) -> AggregateRawData:
        pass


class BaseRepository(ABC):
    def __init__(self, storage: BaseStorage):
        self._storage = storage

    async def create(self) -> Aggregate:
        raw_data = await self._storage.initialize_new_raw_data()
        aggregate_data = self._storage.convert_raw_to_aggregate_data(raw_data=raw_data)
        return self._init_aggregate(data=aggregate_data)

    async def get(self, root_id: UUID) -> Aggregate:
        raw_data = await self._storage.fetch_raw_data(root_id=root_id)
        aggregate_data = self._storage.convert_raw_to_aggregate_data(raw_data=raw_data)
        return self._init_aggregate(data=aggregate_data)

    async def save(self, aggregate: Aggregate) -> None:
        aggregate_data = aggregate.get_data()
        await self._storage.save(data=aggregate_data)

    @abstractmethod
    def _init_aggregate(self, data: AggregateData) -> Aggregate:
        raise NotImplementedError()



from contextlib import AsyncExitStack


class BaseUnit:
    def __init__(self):
        self._context = AsyncExitStack()

    async def __aenter__(self):
        pass

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass

    async def handle_exception(self, exc_type, exc_val, exc_tb):
        pass



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Type, Any

from domain_model.domain_events import BaseDomainEvent, EventsCollector
from domain_model.typing import AggregateData


@dataclass(frozen=True, slots=True)
class BaseCommand:
    command_name = None


class BaseCommandContext:
    pass


class CommandHandler(ABC):
    def __init__(self):
        self._events_collector = None

    def handle(
        self,
        data: AggregateData,
        command: BaseCommand,
        events_collector: EventsCollector,
        context: BaseCommandContext | None,
    ) -> Any | None:
        self._events_collector = events_collector
        result = self._handle(data, command, context)
        self._events_collector = None
        return result

    @abstractmethod
    def _handle(
        self,
        data: AggregateData,
        command: BaseCommand,
        context: BaseCommandContext | None,
    ) -> Any | None:
        pass

    def _add_event(self, event: BaseDomainEvent):
        self._events_collector.add_event(event)


class CommandsStorage:
    def __init__(self, commands_handlers: dict[Type[BaseCommand], CommandHandler]):
        self._commands_handlers = commands_handlers

    def get(self, command: BaseCommand) -> CommandHandler:
        command_type = type(command)
        try:
            return self._commands_handlers[command_type]
        except KeyError:
            raise ValueError(f'Unknown command type: {command_type}')



from domain_model.unit_of_work.units.base import BaseUnit


class BaseDBUnit(BaseUnit):
    def get_db_session(self):
        raise NotImplementedError

    async def rollback(self):
        raise NotImplementedError



class BaseCacheUnit(BaseUnit):
    async def __aenter__(self):
        raise NotImplementedError

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        raise NotImplementedError

    def get_cache(self):
        raise NotImplementedError



from abc import ABC, abstractmethod
from contextlib import  contextmanager
from typing import Any


@contextmanager
def convert_error(error_class: type[Exception]):
    try:
        yield
    except Exception as e:
        raise error_class(*e.args) from e


class BaseUseCase(ABC):
    def __init__(self, error_class: type[Exception]):
        self._error_class = error_class

    @abstractmethod
    async def execute(self, **kwargs) -> Any:
        pass



from typing import Any

from domain_model.commands import BaseCommand, CommandsStorage, CommandHandler, BaseCommandContext
from domain_model.domain_events import BaseDomainEvent, EventsCollector
from domain_model.typing import AggregateData


class Aggregate:
    def __init__(self, commands_storage: CommandsStorage, data: AggregateData, events_collector: EventsCollector = None):
        self._data = data
        self._commands_storage = commands_storage
        self._events_collector: EventsCollector = events_collector or EventsCollector()

    def handle_commands(self, commands: list[BaseCommand]) -> dict[BaseCommand, Any | None]:
        context = self._before_commands(commands=commands)
        result = {}

        for command in commands:
            command_result = self._handle_command(command=command, context=context)
            result[command] =command_result
            self._log(command=command)

        self._after_commands(commands=commands, context=context)
        return result

    def get_data(self) -> AggregateData:
        return self._data

    def get_domain_events(self) -> list[BaseDomainEvent]:
        return self._events_collector.get_events()

    def _before_commands(self, commands: list[BaseCommand]) -> BaseCommandContext | None:
        pass

    def _after_commands(self, commands: list[BaseCommand], context: BaseCommandContext | None) -> None:
        pass

    def _handle_command(self, command: BaseCommand, context: BaseCommandContext) -> Any | None:
        handler = self._pick_handler(command=command)
        return handler.handle(
            data=self._data,
            command=command,
            context=context,
            events_collector=self._events_collector,
        )

    def _pick_handler(self, command: BaseCommand) -> CommandHandler:
        return self._commands_storage.get(command=command)

    def _log(self, command: BaseCommand) -> None:
        pass



from domain_model.unit_of_work.units.cache.base import BaseCacheUnit


class InMemoryCacheUnit(BaseCacheUnit):
    def __init__(self):
        super().__init__()
        self._cache = InMemoryCache()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass

    def get_cache(self):
        return self._cache


class InMemoryCache:
    def __init__(self):
        self.cache = {}

    async def get(self, key):
        return self.cache.get(key)

    async def set(self, key, value, ttl):
        self.cache[key] = value
        return True

    async def delete(self, key):
        if key in self.cache:
            del self.cache[key]
        return True



from typing import Optional, Callable

from sqlalchemy.ext.asyncio import AsyncSession
from domain_model.unit_of_work.units.db.base import BaseDBUnit


class SQLAlchemySessionDBUnit(BaseDBUnit):
    def __init__(self, session_factory: Callable[[], AsyncSession] = None):
        super().__init__()
        self._session_factory = session_factory
        self._session: AsyncSession | None = None

    async def __aenter__(self):
        self._session = self._session_factory()
        await self._context.enter_async_context(self._session.begin())

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            await self.rollback()

        await self._context.aclose()

    def get_db_session(self) -> AsyncSession:
        return self._session

    async def rollback(self):
        return await self._session.rollback()


class SQLAlchemyActiveTransactionSessionDBUnit(BaseDBUnit):
    def __init__(self, session: AsyncSession):
        super().__init__()
        self._session: Optional[AsyncSession] = session

    async def __aenter__(self):
        pass

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass

    def get_db_session(self) -> AsyncSession:
        return self._session

    async def rollback(self):
        pass



from abc import ABC, abstractmethod

from domain_model.domain_events import BaseDomainEvent
from domain_model.typing import AggregateData
from domain_model.unit_of_work.unit_of_work import UnitOfWork


class BaseEventHandler(ABC):
    @abstractmethod
    async def handle(self, data: AggregateData, event: BaseDomainEvent, unit_of_work: UnitOfWork) -> None:
        pass


class DomainEventsHandler(ABC):
    def __init__(self, subscribers: dict[type[BaseDomainEvent], tuple[BaseEventHandler, ...]]):
        self._subscribers = subscribers

    async def handle(self, data: AggregateData, events: list[BaseDomainEvent], unit_of_work: UnitOfWork) -> None:
        for event in events:
            event_handlers = self._subscribers.get(type(event), tuple())
            if not event_handlers:
                continue
            for event_handler in event_handlers:
                await event_handler.handle(data=data, event=event, unit_of_work=unit_of_work)



from uuid import UUID

from domain_model.aggregate import Aggregate
from domain_model.commands import BaseCommand
from domain_model.domain_events import BaseDomainEvent
from domain_model.passport.passport import Passport
from domain_model.typing import AggregateData
from domain_model.unit_of_work.unit_of_work import UnitOfWork


class PassportCommandRunner:
    def __init__(self, passport: Passport):
        self._passport = passport

    async def run_commands(
        self,
        root_id: UUID,
        commands: list[BaseCommand],
        unit_of_work: UnitOfWork,
    ) -> tuple[AggregateData, list[BaseDomainEvent]]:
        repo = self._passport.get_repository(unit_of_work=unit_of_work)
        aggregate: Aggregate = await repo.get(root_id=root_id)
        aggregate.handle_commands(commands=commands)
        return aggregate.get_data(), aggregate.get_domain_events()



from aioredis import Redis

from domain_model.unit_of_work.units.cache import BaseCacheUnit


class RedisCacheUnit(BaseCacheUnit):

    def __init__(self, redis: Redis):
        super().__init__()
        self._redis: Redis = redis
        self._connection = None

    async def __aenter__(self):
        self._connection = await self._context.enter_async_context(self._redis)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._context.aclose()

    def get_cache(self):
        return self._connection




from dataclasses import dataclass


@dataclass(frozen=True, slots=True)
class BaseDomainEvent:
    pass


class EventsCollector:
    def __init__(self):
        self._events = []

    def get_events(self) -> list[BaseDomainEvent]:
        return self._events

    def add_event(self, event: BaseDomainEvent):
        self._events += (event,)

    def __aenter__(self):
        return self

    def __aexit__(self, exc_type, exc_value, traceback):
        pass
